<?php
/**
 * @package     Room reservations
 * @subpackage  Room reservations component
 *
 * @author      Miloslav Schmied
 * @copyright   Copyright (C) 2016 - Ergotep, družstvo invalidů
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 */

// No direct access to this file
defined('_JEXEC') or die('Restricted access');

/**
 * ReservationAdmin Model
 *
 * @since  1.0.0
 */
class JRoomBookingModelReservationAdmin extends JModelAdmin
{
    /**
     * Method to get a table object, load it if necessary.
     *
     * @param   string  $type    The table name. Optional.
     * @param   string  $prefix  The class prefix. Optional.
     * @param   array   $config  Configuration array for model. Optional.
     *
     * @return  JTable  A JTable object
     *
     * @since   1.0.0
     */
    public function getTable($type = 'Reservations', $prefix = 'JROomBookingTable', $config = array())
    {
        return JTable::getInstance($type, $prefix, $config);
    }

    /**
     * Abstract method for getting the form from the model.
     *
     * @param   array $data Data for the form.
     * @param   boolean $loadData True if the form is to load its own data (default case), false if not.
     *
     * @return  mixed  A JForm object on success, false on failure
     *
     * @since   1.0.0
     */
    public function getForm($data = array(), $loadData = true)
    {
        // Get the form.
        $form = $this->loadForm(
            'com_jroombooking.reservation',
            'reservation',
            array(
                'control' => 'jform',
                'load_data' => $loadData
            )
        );

        if (empty($form))
        {
            return false;
        }

        return $form;
    }

    protected function loadFormData()
    {
        // Check the session for previously entered form data.
        $data = JFactory::getApplication()->getUserState(
            'com_jroombooking.edit.reservationadmin.data',
            array()
        );

        if (empty($data))
        {
            $data = $this->getItem();
        }

        return $data;
    }

    /**
     * Method to get the script that have to be included on the form
     *
     * @return string	Script files
     * @since 1.0.0
     */
    public function getScript()
    {
        return 'administrator/components/com_jroombooking/models/forms/reservation.js';
    }

    /**
     * Method to check if it's OK to delete a reservationadmin. Overwrites JModelAdmin::canDelete
     *
     * @since 1.0.0
     */
    protected function canDelete($record)
    {
        if( !empty( $record->id ) )
        {
            return JFactory::getUser()->authorise( "core.delete", "com_jroombooking.reservation." . $record->id );
        }
    }

    /**
     * Metoda pro schvaleni rezervace. Funguje na principu 'zverejneni' polozky.
     *
     * @param mixed $pks Primarni klice polozek.
     * @param int $value Stav schvaleni, 1 = schvalit, 0 = zrusit schvaleni
     * @return bool vysledek provedeni
     * @since 1.0.0
     */
    public function approve(&$pks, $value = 1) {
        $dispatcher = JEventDispatcher::getInstance();
        $user = JFactory::getUser();
        $table = $this->getTable();
        $pks = (array) $pks;

        // Include the plugins for the change of state event.
        JPluginHelper::importPlugin($this->events_map['change_state']);

        // Access checks.
        foreach ($pks as $i => $pk)
        {
            $table->reset();

            if ($table->load($pk))
            {
                if (!$this->canEditState($table))
                {
                    // Prune items that you can't change.
                    unset($pks[$i]);

                    JLog::add(JText::_('JLIB_APPLICATION_ERROR_EDITSTATE_NOT_PERMITTED'), JLog::WARNING, 'jerror');

                    return false;
                }

                // If the table is checked out by another user, drop it and report to the user trying to change its state.
                if (property_exists($table, 'checked_out') && $table->checked_out && ($table->checked_out != $user->id))
                {
                    JLog::add(JText::_('JLIB_APPLICATION_ERROR_CHECKIN_USER_MISMATCH'), JLog::WARNING, 'jerror');

                    // Prune items that you can't change.
                    unset($pks[$i]);

                    return false;
                }
            }
        }

        // Attempt to change the state of the records.
        if (!$table->approve($pks, $value, $user->get('id')))
        {
            $this->setError($table->getError());

            return false;
        }

        $context = $this->option . '.' . $this->name;

        // Trigger the change state event.
        $result = $dispatcher->trigger($this->event_change_state, array($context, $pks, $value));

        if (in_array(false, $result, true))
        {
            $this->setError($table->getError());

            return false;
        }

        // Clear the component's cache
        $this->cleanCache();

        return true;
    }

    /**
     * Metoda prekryvajici metodu 'save' predka. Datum se pro ulozeni do DB musi prevest do spravneho formatu.
     *
     * @param array $data Data pro ulozeni do DB
     * @return bool vysledek ulozeni
     * @since 1.0.0
     */
    public function save($data)
    {
        //odstranenni mezer z datumu, s mezerami selze parsovani
        $startDate = str_replace(' ', '', $data['start_date']);
        $endDate = str_replace(' ', '', $data['end_date']);
        $startDate = new JDate($startDate);
        $data['start_date'] = $startDate->toSql();
        $endDate = new JDate($endDate);
        $data['end_date'] = $endDate->toSql();

        return parent::save($data);
    }

    /**
     * Metoda pro validaci dat formulare. Kapacita se validuje podle maximalni kapacity pokoje.
     * Datum rezervace se validuje podle jiz schvalenych rezervaci pro stejny pokoj.
     *
     * @param JForm $form Formular
     * @param array $data Data z formulare
     * @param null $group
     * @return bool|mixed
     * @since 1.0.0
     */
    public function validate($form, $data, $group = null)
    {
        // validace kapacity
        if (!$this->validateCapacity($data['room_id'], $data['capacity'])) {
            return false;
        }

        // validace terminu
        if (!$this->validateDates($data['id'], $data['room_id'], $data['start_date'], $data['end_date'])) {
            return false;
        }

        return parent::validate($form, $data, $group);
    }

    /**
     * Validuje datum rezervace. Datum nesmi kolidovat s jiz schvalenou regisstraci pro stejny pokoj.
     *
     * @param int $id ID rezervace
     * @param int $roomId ID pokoje
     * @param string $startDate Datum zacatku rezervace
     * @param string $endDate Datum konce rezervace
     * @return bool Vysledek validace
     * @since 1.0.0
     */
    private function validateDates($id, $roomId, $startDate, $endDate) {
        // prevod datumu do spravneo formatu pro DB
        $startDate = new JDate($startDate);
        $endDate = new JDate($endDate);

        // datum zacatku musi byt drive nez datum konce
        if ($startDate >= $endDate) {
            JFactory::getApplication()->enqueueMessage(JText::_('COM_JROOMBOOKING_RESERVE_START_DATE_LESS_OR_EQUAL_END_DATE'), 'error');
            return false;
        }

        // kontrola kolize s jiz schvalenymi rezervacemi
        $db = JFactory::getDbo();
        $query = $db->getQuery(true);
        $query->select($db->qn(array('id', 'start_date', 'end_date')))
            ->from('#__jroombooking_reservations')
            ->where($db->qn('approved') . ' = ' . $db->q(1) . ' AND ' . $db->qn('room_id') . ' = ' . $db->q($roomId)
                . ' AND ' . $db->qn('id') . ' <> ' . $db->q($id) . ' AND ((' . $db->qn('start_date') . ' >= ' . $db->q($startDate->toSql())
                . ' AND ' . $db->qn('end_date') . ' <= ' . $db->q($endDate->toSql()) . ') OR (' . $db->qn('end_date') . ' >= ' . $db->q($startDate->toSql())
                . ' AND ' . $db->qn('start_date') . ' <= '. $db->q($endDate->toSql()) .'))');
        $db->setQuery($query);
        $result = $db->loadObjectList();
        if (sizeof($result) > 0) {
            JFactory::getApplication()->enqueueMessage(JText::_('COM_JROOMBOOKING_RESERVE_ROOM_OCCUPIED'), 'error');
            JFactory::getApplication()->enqueueMessage(JText::_('COM_JROOMBOOKING_RESERVATIONS_COLLISIONS'), 'error');
            foreach ($result as $collision) {
                $startDate = new JDate($collision->start_date);
                $endDate = new JDate($collision->end_date);
                $format = JText::_('COM_JROOMBOOKING_RESERVATIONS_SHORT_DATE_FORMAT');
                $colisionInfo = JText::_('COM_JROOMBOOKING_RESERVATIONS_ID') . ': ' . $collision->id . ' (' . $startDate->format($format) . ' - ' . $endDate->format($format) . ')';
                JFactory::getApplication()->enqueueMessage($colisionInfo, 'error');
            }
            return false;
        }
        return true;
    }

    /**
     * Metoda pro validaci rezervovane kapacity. Rezervovana kapacita nesmi byt vetsi nez maximalni kapacita pokoje.
     *
     * @param int $roomId
     * @param int $capacity
     * @return bool
     * @since 1.0.0
     */
    private function validateCapacity($roomId, $capacity) {
        $model =& $this->getInstance('JRoomBooking', 'JRoomBookingModel');
        $room = $model->getItem($roomId);
        if ($capacity > $room->max_capacity) {
            JFactory::getApplication()->enqueueMessage(JText::sprintf('COM_JROOMBOOKING_RESERVE_CAPACITY_EXCEEDED', $room->max_capacity), 'error');
            return false;
        }
        return true;
    }

    /**
     * Ziska data rezervace pro zadane ID rezervace.
     *
     * @param int $id ID rezervace
     * @return mixed Objekt s daty
     * @since 1.0.0
     */
    public function getReservation($id) {
        $db = JFactory::getDbo();
        $query = $db->getQuery(true);
        $query->select($db->qn(array('a.id', 'a.email', 'a.firstname', 'a.surname', 'a.start_date', 'a.end_date', 'a.capacity', 'b.name')))
            ->from($db->qn('#__jroombooking_reservations') . ' AS a')
            ->leftJoin($db->qn('#__jroombooking_rooms') . ' AS b ON (' . $db->qn('a.room_id') . ' = ' . $db->qn('b.id') . ')')
            ->where($db->qn('a.id') . ' = ' . $db->q($id));
        $db->setQuery($query);
        $email = $db->loadObject();
        return $email;
    }
}