<?php
/**
 * @package     Room reservations
 * @subpackage  Room reservations component
 *
 * @author      Miloslav Schmied
 * @copyright   Copyright (C) 2016 - Ergotep, družstvo invalidů
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 */

// No direct access
defined('_JEXEC') or die('Restricted access');

/**
 * Reservations Table class
 *
 * @since 1.0.0
 */
class JRoomBookingTableReservations extends JTable
{
    /**
     * Constructor
     *
     * @param JDatabaseDriver  &$db  A database connector object
     * @since 1.0.0
     */
    function __construct(&$db)
    {
        parent::__construct('#__jroombooking_reservations', 'id', $db);
    }

    public function bind($src, $ignore = array())
    {
        //nastavi sloupec date_modified na aktualni cas, kdyz dojde k editaci zaznamu
        $now = new JDate();
        $this->date_modified = $now->toSql();

        return parent::bind($src, $ignore);
    }

    /**
     * Metoda pro schvaleni rezervace.
     *
     * @param null $pks Primarni klice
     * @param int $state Stav schvaleno/neschvaleno
     * @param int $userId ID uzivatele
     * @return bool Vysledek schvaleni
     * @since 1.0.0
     */
    public function approve($pks = null, $state = 1, $userId = 0) {
        // Sanitize input
        $userId = (int) $userId;
        $state  = (int) $state;

        if (!is_null($pks))
        {
            if (!is_array($pks))
            {
                $pks = array($pks);
            }

            foreach ($pks as $key => $pk)
            {
                if (!is_array($pk))
                {
                    $pks[$key] = array($this->_tbl_key => $pk);
                }
            }
        }

        // If there are no primary keys set check to see if the instance key is set.
        if (empty($pks))
        {
            $pk = array();

            foreach ($this->_tbl_keys AS $key)
            {
                if ($this->$key)
                {
                    $pk[$key] = $this->$key;
                }
                // We don't have a full primary key - return false
                else
                {
                    $this->setError(JText::_('JLIB_DATABASE_ERROR_NO_ROWS_SELECTED'));

                    return false;
                }
            }

            $pks = array($pk);
        }

        //kontrola, jestli termin rezervace nekoliduje s jiz schvalenymi rezervacemi
        if ($state == 1) {
            foreach ($pks as $key=>$pk) {
                $valid = $this->validateDate($pk[$this->getKeyName()]);
                if (!sizeof($valid) == 0) {
                    JFactory::getApplication()->enqueueMessage(JText::_('COM_JROOMBOOKING_RESERVATIONS_APPROVE_DATE_ERROR'), 'error');
                    JFactory::getApplication()->enqueueMessage(JText::_('COM_JROOMBOOKING_RESERVATIONS_COLLISIONS'), 'error');
                    foreach ($valid as $collision) {
                        $startDate = new JDate($collision->start_date);
                        $endDate = new JDate($collision->end_date);
                        $format = JText::_('COM_JROOMBOOKING_RESERVATIONS_SHORT_DATE_FORMAT');
                        $colisionInfo = JText::_('COM_JROOMBOOKING_RESERVATIONS_ID') . ': ' . $collision->id . ' (' . $startDate->format($format) . ' - ' . $endDate->format($format) . ')';
                        JFactory::getApplication()->enqueueMessage($colisionInfo, 'error');
                    }
                    return false;
                }
            }
        }


        $approvedField = $this->getColumnAlias('approved');
        $checkedOutField = $this->getColumnAlias('checked_out');

        foreach ($pks as $pk)
        {
            // Update the publishing state for rows with the given primary keys.
            $query = $this->_db->getQuery(true)
                ->update($this->_tbl)
                ->set($this->_db->quoteName($approvedField) . ' = ' . (int) $state);

            // aktualizace sloupce 'date_approved_modified' na aktualni datum a cas
            $nowDate = new JDate();
            $nowDate = $this->_db->quote($nowDate->toSql());
            $query->set($this->_db->quoteName($this->getColumnAlias('date_approved_modified')) . ' = ' . $nowDate);

            // If publishing, set published date/time if not previously set
            if ($state && property_exists($this, 'publish_up') && (int) $this->publish_up == 0)
            {
                $query->set($this->_db->quoteName($this->getColumnAlias('publish_up')) . ' = ' . $nowDate);
            }

            // Determine if there is checkin support for the table.
            if (property_exists($this, 'checked_out') || property_exists($this, 'checked_out_time'))
            {
                $query->where('(' . $this->_db->quoteName($checkedOutField) . ' = 0 OR ' . $this->_db->quoteName($checkedOutField) . ' = ' . (int) $userId . ')');
                $checkin = true;
            }
            else
            {
                $checkin = false;
            }

            // Build the WHERE clause for the primary keys.
            $this->appendPrimaryKeys($query, $pk);

            $this->_db->setQuery($query);

            try
            {
                $this->_db->execute();
            }
            catch (RuntimeException $e)
            {
                $this->setError($e->getMessage());

                return false;
            }

            // If checkin is supported and all rows were adjusted, check them in.
            if ($checkin && (count($pks) == $this->_db->getAffectedRows()))
            {
                $this->checkin($pk);
            }

            // If the JTable instance value is in the list of primary keys that were set, set the instance.
            $ours = true;

            foreach ($this->_tbl_keys AS $key)
            {
                if ($this->$key != $pk[$key])
                {
                    $ours = false;
                }
            }

            if ($ours)
            {
                $this->$approvedField = $state;
            }
        }

        $this->setError('');

        return true;
    }

    private function validateDate($pk) {
        // znamenko pro porovnani konce rezervace podle toho, zda je umozneno vytvorit novou rezervaci v terminu konce predchozi rezervace
        $compareSign = '>=';
        if(JComponentHelper::getParams('com_jroombooking')->get('same_day_reservation', '1'))
        {
          $compareSign = '>';
        }
        $reservation = $this->getReservation($pk);
        $startDate = $reservation->start_date;
        $endDate = $reservation->end_date;
        $id = $reservation->room_id;

        $query = $this->_db->getQuery(true);
        $query->select($this->_db->qn(array('id', 'start_date', 'end_date')))
            ->from('#__jroombooking_reservations')
            ->where($this->_db->qn('approved') . ' = ' . $this->_db->q(1) . ' AND ' . $this->_db->qn('room_id') . ' = ' . $this->_db->q($id)
                . ' AND ((' . $this->_db->qn('start_date') . ' >= ' . $this->_db->q($startDate)
                . ' AND ' . $this->_db->qn('end_date') . ' <= ' . $this->_db->q($endDate) . ') OR (' . $this->_db->qn('end_date') . ' '.$compareSign.' ' . $this->_db->q($startDate)
                . ' AND ' . $this->_db->qn('start_date') . ' <= '. $this->_db->q($endDate) .'))');
        $this->_db->setQuery($query);
        $result = $this->_db->loadObjectList();
        return $result;
    }

    private function getReservation($id) {
        $query = $this->_db->getQuery(true);
        $query->select($this->_db->qn(array('id', 'room_id', 'start_date', 'end_date')))
        ->from($this->_db->qn('#__jroombooking_reservations'))
            ->where($this->_db->qn('id') . ' = ' . $this->_db->q($id));
        $this->_db->setQuery($query);
        $reservation = $this->_db->loadObject();
        return $reservation;
    }
}