<?php
/**
 * Library for Joomla for creating thumbnails of images
 *
 * @package Mavik Thumb
 * @version 1.0
 * @author Vitaliy Marenkov <admin@mavik.com.ua>
 * @copyright 2012 Vitaliy Marenkov
 * @license GNU General Public License version 2 or later; see LICENSE.txt
 */
defined('_JEXEC') or die;

jimport('mavik.thumb.graphiclibrary.abstract');
jimport('joomla.filesystem.file');

/**
 * Adapter for Graphic Library GD2
 */
class MavikThumbGraphicLibraryGd2 extends MavikThumbGraphicLibrary {

    public function __construct()
    {
        $this->checkRequirements();
    }

    /**
     * Check the server requirements
     */
    protected function checkRequirements()
    {
        // Check version of GD
        if (!function_exists('imagecreatetruecolor')) {
            throw new Exception(JText::_('Library mAvik Thumb needs library GD2'), MavikThumbGenerator::ERROR_LIBRARY_IS_MISSING);
        }
    }

    /**
     * @param MavikThumbInfo $info
     * @param int $x
     * @param int $y
     * @param int $widht
     * @param int $height
     */
    public function createThumbnail(MavikThumbInfo $info, $x, $y, $widht, $height)
    {
        $orig = $this->createOrigin($info);
        $thumb = imagecreatetruecolor($info->thumbnail->realWidth, $info->thumbnail->realHeight);
        $colorsCount = $this->processTransparent($orig, $thumb, $info);
        $this->createThumbnailFile($thumb, $orig, $info, $colorsCount, $x, $y, $widht, $height);
    }

    /**
     * @param MavikThumbInfo $info
     * @return resource
     * @throws Exception
     */
    private function createOrigin(MavikThumbInfo $info)
    {
        switch ($info->original->type)
        {
            case 'image/jpeg':
                return imagecreatefromjpeg($info->original->path);
            case 'image/png':
                return imagecreatefrompng($info->original->path);
            case 'image/gif':
                return imagecreatefromgif($info->original->path);
            default:
                throw new Exception(JText::sprintf('Unsupported type of image', $info->original->type), MavikThumbGenerator::ERROR_UNSUPPORTED_TYPE);
        }
    }

    /**
     * 
     * @param resource $orig
     * @param resource $thumb
     * @param MavikThumbInfo $info
     * @return int
     */
    private function processTransparent($orig, $thumb, MavikThumbInfo $info)
    {
        $colorsCount = imagecolorstotal($orig);
        if ($info->original->type == 'image/png' || $info->original->type == 'image/gif') {
            $transparentIndex = imagecolortransparent($orig);
            if ($transparentIndex >= 0 && $transparentIndex < $colorsCount)
            {
                // without alpha-chanel
                $transparentRgb = imagecolorsforindex($orig, $transparentIndex);
                imagecolordeallocate($orig, $transparentIndex);
                $transparentIndex = imagecolorallocate($orig, $transparentRgb['red'], $transparentRgb['green'], $transparentRgb['blue']);
                imagefilledrectangle( $thumb, 0, 0, $info->thumbnail->realWidth, $info->thumbnail->realHeight, $transparentIndex );
                imagecolortransparent($thumb, $transparentIndex);
            } elseif ($info->original->type == 'image/png') {
                // with alpha-chanel
                imagealphablending($thumb, false);
                imagesavealpha($thumb, true);
                $transparent = imagecolorallocatealpha($thumb, 255, 255, 255, 127);
                imagefilledrectangle($thumb, 0, 0, $info->thumbnail->realWidth, $info->thumbnail->realHeight, $transparent);
            }
        }

        return $colorsCount;
    }

    /**
     * @param resource $thumb
     * @param resource $orig
     * @param MavikThumbInfo $info
     * @param int $colorsCount
     * @param int $x
     * @param int $y
     * @param int $widht
     * @param int $height
     * @throws Exception
     */
    private function createThumbnailFile($thumb, $orig, MavikThumbInfo $info, $colorsCount, $x, $y, $widht, $height)
    {
        imagecopyresampled($thumb, $orig, 0, 0, $x, $y, $info->thumbnail->realWidth, $info->thumbnail->realHeight, $widht, $height);
        if ($colorsCount) {
            imagetruecolortopalette($thumb, false, $colorsCount);
        }
        ob_start();
        switch ($info->original->type)
        {
            case 'image/jpeg':
                    $result = imagejpeg($thumb, null, $this->quality);
                    break;
            case 'image/png':
                    $result = imagepng($thumb, null, 9);
                    break;
            case 'image/gif':
                    $result = imagegif($thumb, null);
        }
        JFile::write($info->thumbnail->path, ob_get_contents());
        ob_end_clean();

        if(!$result) {
            throw new Exception(JText::sprintf('Can\'t create thumbnail for', $info->original->path), MavikThumbGenerator::ERROR_FILE_CREATION);
        }

        imagedestroy($orig);
        imagedestroy($thumb);
    }
}